/* Generated by generate_rules.sh; do not edit */
static struct { const char *uses; const char *flavor; } static_flavors_[] = {
	{ "emacs", "canna" },
	{ "emacs", "devel_full" },
	{ "emacs", "devel_nox" },
	{ "emacs", "full" },
	{ "emacs", "nox" },
	{ "lazarus", "gtk2" },
	{ "lazarus", "qt5" },
	{ "lua", "lua51" },
	{ "lua", "lua52" },
	{ "lua", "lua53" },
	{ "lua", "lua54" },
	{ "php", "php73" },
	{ "php", "php74" },
	{ "php", "php80" },
	{ "python", "py27" },
	{ "python", "py36" },
	{ "python", "py37" },
	{ "python", "py38" },
	{ "python", "py39" },
};
static const char *known_architectures_[] = {
	"aarch64",
	"amd64",
	"arm",
	"armv6",
	"armv7",
	"i386",
	"mips",
	"mips64",
	"mips64el",
	"mips64elhf",
	"mips64hf",
	"mipsel",
	"mipselhf",
	"mipshf",
	"mipsn32",
	"powerpc",
	"powerpc64",
	"powerpc64le",
	"powerpcspe",
	"riscv64",
	"riscv64sf",
	"sparc64",
};
